import { ClassOf } from './immutable-interfaces';
/**
 * Declares a class as immutable.
 * This creates a clone of object to a specified depth when a method is called.
 * If mutability checking is enabled (`immutableSettings()`), calling a method throws an exception
 * if the method modifies values of references in the object's properties or the method arguments.
 *
 * @example
 *     addToList(item: Item) {
 *         // These two will throw exceptions:
 *         this.list.push(item);
 *         this.list[this.list.length++] = item;
 *
 *         // These create a new reference and work:
 *         this.list = [...this.list, newItem];
 *         this.list = this.list.concat(newItem);
 *     }
 *
 * Changing the depth will clone object properties to the depth specified:
 *     @Immutable()
 *     class Example {
 *         @CloneDepth(<the depth you want to use>)
 *         state = { list: [{ id: 1, name: 'one' }] };
 *         addToList(item: { id: number, name: string }) {
 *             // with depth 0:
 *             this.state = { ...this.state, list: [...this.state.list, item] };
 *             // with depth 1:
 *             this.state.list = [...this.state.list, item];
 *             // with depth 2:
 *             this.state.list.push(item);
 *         };
 *     }
 */
export declare function Immutable(): <T, C extends ClassOf<T>>(target: C) => C;
/** Declares a class as immutable. */
export declare function Immutable<T, C extends ClassOf<T>>(target: C): C;
