"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ImmutableStateStore = void 0;
var observe_immutable_1 = require("./observe-immutable");
var utils_1 = require("./utils");
/**
 * Implements an observable store with separate state action branches.
 */
var ImmutableStateStore = /** @class */ (function () {
    function ImmutableStateStore(actions) {
        this.actions = actions;
        this.observers = [];
        this.callObservers = [];
        this.subscriptions = [];
        this.combineInitialState();
        this.observeActionMethodCalls();
    }
    Object.defineProperty(ImmutableStateStore.prototype, "state", {
        get: function () {
            return this.currentState;
        },
        enumerable: false,
        configurable: true
    });
    /** Remove all observers and free resources. */
    ImmutableStateStore.prototype.destroy = function () {
        this.observers.splice(0, this.observers.length);
        this.callObservers.splice(0, this.callObservers.length);
        this.subscriptions.forEach(function (s) { return s.unsubscribe(); });
    };
    ImmutableStateStore.prototype.observeCalls = function (observableClass) {
        return observe_immutable_1.createObservable(this.callObservers, observableClass);
    };
    ImmutableStateStore.prototype.observeState = function (observableClass) {
        return observe_immutable_1.createObservable(this.observers, observableClass);
    };
    /** Fully replace the application state (e.g. for undo/redo) */
    ImmutableStateStore.prototype.replaceState = function (newState) {
        var changes = {};
        for (var _i = 0, _a = utils_1.objectKeys(newState); _i < _a.length; _i++) {
            var key = _a[_i];
            if (this.currentState[key] !== newState[key]) {
                changes[key] = {
                    oldValue: this.currentState[key],
                    newValue: newState[key]
                };
            }
        }
        if (Object.keys(changes).length > 0) {
            this.propertiesChanged(undefined, changes);
            this.emitChangedState();
        }
    };
    /** Should only be used in testing. Use action methods to change the application state. */
    ImmutableStateStore.prototype.setStateForTesting = function (newState) {
        var changes = {};
        for (var _i = 0, _a = utils_1.objectKeys(newState); _i < _a.length; _i++) {
            var key = _a[_i];
            changes[key] = {
                oldValue: this.currentState[key],
                newValue: newState[key]
            };
        }
        this.propertiesChanged(undefined, changes);
        this.emitChangedState();
    };
    ImmutableStateStore.prototype.combineInitialState = function () {
        // Combine initial state properties of all branches into one object.
        var initialState = {};
        for (var branchName in this.actions) {
            var actionBranch = this.actions[branchName];
            for (var _i = 0, _a = utils_1.objectKeys(actionBranch); _i < _a.length; _i++) {
                var propKey = _a[_i];
                if (actionBranch[propKey] !== undefined || !(propKey in initialState)) {
                    initialState[propKey] = actionBranch[propKey];
                }
            }
        }
        this.currentState = initialState;
        // Apply initial state to all action branches
        for (var branchName in this.actions) {
            var actionBranch = this.actions[branchName];
            for (var _b = 0, _c = utils_1.objectKeys(actionBranch); _b < _c.length; _b++) {
                var propKey = _c[_b];
                actionBranch[propKey] = initialState[propKey];
            }
        }
    };
    ImmutableStateStore.prototype.observeActionMethodCalls = function () {
        var _this = this;
        // Observe all method calls of any action branch
        for (var branchName in this.actions) {
            var actionBranch = this.actions[branchName];
            var sub = observe_immutable_1.observeImmutable(actionBranch).subscribe(function (call) {
                var stateBefore = _this.currentState;
                if (call.changes) {
                    _this.propertiesChanged(call.instance, call.changes);
                    _this.emitChangedState();
                }
                _this.emitMethodCall(call, stateBefore);
            });
            this.subscriptions.push(sub);
        }
    };
    // Apply changes to this.currentState and all action branches that use the changed property
    ImmutableStateStore.prototype.propertiesChanged = function (instance, changes) {
        var newState = __assign({}, this.state);
        for (var changedProp in changes) {
            var newValue = changes[changedProp].newValue;
            newState[changedProp] = newValue;
            for (var branchToUpdateName in this.actions) {
                var branchToUpdate = this.actions[branchToUpdateName];
                if (branchToUpdate != instance && utils_1.hasOwnProperty(branchToUpdate, changedProp)) {
                    branchToUpdate[changedProp] = newValue;
                }
            }
        }
        this.currentState = newState;
    };
    ImmutableStateStore.prototype.emitChangedState = function () {
        for (var _i = 0, _a = this.observers; _i < _a.length; _i++) {
            var observer = _a[_i];
            observer(this.currentState);
        }
    };
    ImmutableStateStore.prototype.emitMethodCall = function (actionMethodCall, stateBefore) {
        var _this = this;
        if (this.callObservers.length) {
            // Emit a tracked method call on "this".
            var actionName = Object.keys(this.actions)
                .filter(function (propertyKey) { return _this.actions[propertyKey] === actionMethodCall.instance; })[0];
            var callToEmit = {
                instance: this,
                methodName: actionName + '.' + actionMethodCall.methodName,
                arguments: actionMethodCall.arguments,
                returnValue: actionMethodCall.returnValue,
                callDuration: actionMethodCall.callDuration,
                changes: actionMethodCall.changes,
                oldProperties: stateBefore,
                newProperties: this.currentState,
            };
            for (var _i = 0, _a = this.callObservers; _i < _a.length; _i++) {
                var observer = _a[_i];
                observer(callToEmit);
            }
        }
    };
    ImmutableStateStore.prototype.handleSubscription = function (subscriber) {
        var _this = this;
        var callback = subscriber.next
            ? function (newState) { return subscriber.next(newState); }
            : subscriber;
        this.observers.push(callback);
        return {
            unsubscribe: function () {
                var index;
                while ((index = _this.observers.indexOf(callback)) >= 0) {
                    _this.observers.splice(index, 1);
                }
            }
        };
    };
    return ImmutableStateStore;
}());
exports.ImmutableStateStore = ImmutableStateStore;

//# sourceMappingURL=immutable-state-store.js.map
