"use strict";
var __spreadArrays = (this && this.__spreadArrays) || function () {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.logMethodCallToConsole = void 0;
var deep_clone_1 = require("./deep-clone");
var utils_1 = require("./utils");
var immutable_state_store_1 = require("./immutable-state-store");
/** Logger that logs a method call to the console. Works in browsers and node. */
function logMethodCallToConsole(call) {
    /** Internet Explorer only sets window.console while the developer tools are opened. */
    if (typeof console !== 'object' || !console || !console.log || !call) {
        return;
    }
    var originalMethod = getOriginalMethod(call.instance, call.methodName);
    var support = determineConsoleFeaturesSupportedByEnvironment();
    var timeColor = call.callDuration > 200 ? 'red' : (call.callDuration > 50 ? 'orange' : 'green');
    var changedBranches = Object.keys(call.changes || {});
    var duration = (call.callDuration || 0).toFixed(2);
    if (support.styledGroups) {
        // Full-featured browser output, only supported by chrome (march 2017).
        console.groupCollapsed(call.methodName + " %c(" + duration + ")", "color: " + timeColor);
        console.log("time: %c" + duration + "ms", "color: " + timeColor);
        console.log('method: ', originalMethod);
        if (call.arguments && call.arguments.length) {
            console.log.apply(console, __spreadArrays(['arguments: '], deep_clone_1.deepClone(call.arguments, 5)));
        }
        else {
            console.log('arguments: %c<none>', 'color: #999');
        }
        if (call.returnValue !== undefined) {
            console.log('return value: ', call.returnValue);
        }
        console.groupCollapsed('%cchanged state branches: %c' + changedBranches.join(', '), 'font-weight: normal', 'font-weight: bold');
        for (var _i = 0, changedBranches_1 = changedBranches; _i < changedBranches_1.length; _i++) {
            var key = changedBranches_1[_i];
            console.log("%c" + key + " changed from ", 'font-style: italic', call.changes[key].oldValue, ' to ', call.changes[key].newValue);
        }
        console.groupEnd();
        console.log('state before: ', call.oldProperties);
        console.log('state after: ', call.newProperties);
        console.groupEnd();
    }
    else if (support.expandableObjects) {
        // Firefox, IE, Edge, Safari, etc
        var outputObject = {
            time: call.callDuration.toFixed(2) + 'ms',
            method: originalMethod,
            arguments: call.arguments && call.arguments.length ? call.arguments : '<none>',
            returnValue: call.returnValue,
            changedBranches: call.changes,
            stateBefore: call.oldProperties,
            stateAfter: call.newProperties
        };
        if (support.styledOutput) {
            console.log(call.methodName + " %c(" + duration + ")%c - %o", "color: " + timeColor, 'color: black', outputObject);
        }
        else {
            console.log(call.methodName + " (" + duration + ") - ", outputObject);
        }
    }
    else if (support.terminalEscapeSequences) {
        // Node when run from command-line
        console.log(ansiSequence('underline') + call.methodName + ansiSequence(timeColor) + ' (' + duration + ')' + ansiSequence('reset'));
        console.log("    time: " + ansiSequence(timeColor) + duration + "ms" + ansiSequence('reset'));
        if (call.arguments && call.arguments.length) {
            console.log('    arguments: ');
            for (var _a = 0, _b = call.arguments; _a < _b.length; _a++) {
                var arg = _b[_a];
                console.log('    - ' + JSON.stringify(arg, null, 2).replace(/\n/g, '\n    '));
            }
        }
        else {
            console.log("    arguments: " + ansiSequence('gray') + "<none>" + ansiSequence('reset'));
        }
        if (call.returnValue !== undefined) {
            console.log('    return value: ', call.returnValue);
        }
        console.log('    changed state branches: ' + changedBranches.join(', '));
    }
    else {
        // Node with debugger attached, weird browsers
        console.log(call.methodName + " (" + duration + ")");
    }
}
exports.logMethodCallToConsole = logMethodCallToConsole;
function getOriginalMethod(instance, methodName) {
    var isStore = instance instanceof immutable_state_store_1.ImmutableStateStore;
    var methodParts = methodName.split('.');
    var target = isStore ? instance.actions && instance.actions[methodParts[0]] : instance;
    var metadata = target && utils_1.getImmutableClassMetadata(Object.getPrototypeOf(target).constructor);
    var originalPrototype = metadata && metadata.originalClass.prototype;
    var propertyKeys = isStore ? methodParts.slice(1) : methodParts;
    var obj = originalPrototype;
    for (var _i = 0, propertyKeys_1 = propertyKeys; _i < propertyKeys_1.length; _i++) {
        var part = propertyKeys_1[_i];
        obj = obj && obj[part];
    }
    return obj || undefined;
}
function ansiSequence(color) {
    switch (color) {
        case 'red': return '\x1B[31m';
        case 'green': return '\x1B[0;31m';
        case 'gray': return '\x1B[0;30;1m';
        case 'orange':
        case 'yellow':
            return '\x1B[0;33m';
        case 'underline':
            return '\x1B[0;4m';
        case 'black':
        case 'reset':
        default:
            return '\x1B[0m';
    }
}
var support = undefined;
function determineConsoleFeaturesSupportedByEnvironment() {
    if (support) {
        return support;
    }
    else if (typeof console !== 'object') {
        return support = {
            collapsedGroups: false,
            expandableObjects: false,
            groups: false,
            styledGroups: false,
            styledOutput: false,
            terminalEscapeSequences: false
        };
    }
    // Unfortunately, console features can only be detected via user agent sniffing :-/
    var userAgent = typeof navigator === 'object' ? navigator.userAgent : '';
    var isChrome = /Chrome/.test(userAgent) && /Google Inc/.test(navigator.vendor) && !/OPR/.test(userAgent);
    var isIE = typeof document === 'object' && !!document.documentMode;
    var isEdge = / Edge\//.test(userAgent);
    var isFirefox = /firefox/i.test(userAgent);
    var isFirebug = !isIE && console.log.toString().indexOf('apply') !== -1;
    var isNode = typeof window === 'undefined' && typeof global === 'object';
    var isTTY = isNode && (typeof process === 'object' && process.stdout && process.stdout.isTTY);
    var styleable = !isIE && !isEdge && !isNode && (!isFirefox || isFirebug);
    return support = {
        collapsedGroups: !!console.groupCollapsed && !isIE && !isEdge && (!isFirefox || isFirebug),
        expandableObjects: !isNode || (typeof v8debug !== 'undefined'),
        groups: !isIE && !!console.group,
        styledGroups: isChrome && styleable,
        styledOutput: styleable,
        terminalEscapeSequences: isNode && isTTY && (typeof v8debug === 'undefined')
    };
}

//# sourceMappingURL=log-call-to-console.js.map
