import { deepClone } from './deep-clone';
import { flatEqual } from './flat-equal';
import { functionMutatesInput } from './function-mutates-input';
import { globalSettings } from './immutable-settings';
import { MethodNotImmutableError } from './method-not-immutable-error';
import { deepFreeze } from './deep-freeze';
import { getFunctionName, getImmutableClassMetadata, getInstanceMetadata, hasOwnProperty, objectAssign, objectCreate, objectDefineProperty, objectKeys, setImmutableClassMetadata, setInitialInstanceMetadata } from './utils';
let constructing = false;
export function Immutable(originalClass) {
    if (originalClass) {
        return createImmutableClass(originalClass);
    }
    else {
        return Immutable;
    }
}
/** @internal */
export function createImmutableClass(originalClass) {
    const originalMetadata = getImmutableClassMetadata(originalClass);
    const metadata = {
        cloneDepth: originalMetadata && originalMetadata.cloneDepth || {},
        originalClass,
        settings: globalSettings,
    };
    // Constructor that will be called on `new Class()` instead of the original constructor.
    function mappedConstructor(...args) {
        constructing = true;
        let instance = new originalClass(...args);
        constructing = false;
        for (const key of Object.getOwnPropertyNames(instance)) {
            const descriptor = Object.getOwnPropertyDescriptor(instance, key);
            objectDefineProperty(this, key, descriptor);
        }
        8;
        setInitialInstanceMetadata(this);
        if (metadata.settings.deepFreeze) {
            for (let key of objectKeys(this)) {
                const propertyValue = this[key];
                if (typeof propertyValue === 'object' && propertyValue) {
                    deepFreeze(propertyValue);
                }
            }
        }
    }
    const className = getFunctionName(originalClass);
    const mappedClass = createNamedFunction(className, mappedConstructor);
    const originalPrototype = originalClass.prototype;
    const mappedPrototype = mappedClass.prototype = objectCreate(originalClass.prototype);
    // Map methods of the original class
    for (let propertyKey of Object.getOwnPropertyNames(originalPrototype)) {
        let method = originalPrototype[propertyKey];
        if (propertyKey !== 'constructor' && typeof method === 'function') {
            mappedPrototype[propertyKey] = createMethodWrapper(method, propertyKey, metadata);
        }
    }
    objectDefineProperty(mappedPrototype, 'constructor', {
        configurable: true,
        enumerable: false,
        value: mappedClass,
        writable: true
    });
    // Copy static methods and properties
    for (let staticKey of Object.getOwnPropertyNames(originalClass)) {
        if (!hasOwnProperty(mappedClass, staticKey)) {
            const descriptor = Object.getOwnPropertyDescriptor(originalClass, staticKey);
            objectDefineProperty(mappedClass, staticKey, descriptor);
        }
    }
    // Define immutable-specific metadata
    setImmutableClassMetadata(mappedClass, metadata);
    setImmutableClassMetadata(mappedPrototype, metadata);
    return mappedClass;
}
/** Creates a new function with the passed name. */
function createNamedFunction(name, realImplementation) {
    name = name.replace(/[+-.,;()\{\}\[\]]+/g, '_');
    const creator = new Function('fn', `return function ${name}() { return fn.apply(this, arguments); }`);
    return creator(realImplementation);
}
function createMethodWrapper(originalMethod, methodName, classMetadata) {
    let wrapped = function immutableWrappedMethod(...args) {
        const { cloneDepth, originalClass, settings } = classMetadata;
        const instanceMetadata = (getInstanceMetadata(this) || {});
        const originalProperties = {};
        let returnValue;
        let timeBefore;
        let callDuration = -1;
        // Only check mutability when enabled (e.g. in development)
        const runAndCheckChanges = settings.checkMutability
            ? functionMutatesInput
            : ((callback) => { callback(); return false; });
        const mutations = runAndCheckChanges(() => {
            for (let key of objectKeys(this)) {
                originalProperties[key] = this[key];
                const propertyCloneDepth = cloneDepth[key];
                if (propertyCloneDepth > 0) {
                    this[key] = deepClone(this[key], propertyCloneDepth - 1);
                }
            }
            timeBefore = new Date();
            const timestampBefore = getTimestamp();
            instanceMetadata.callDepth++;
            try {
                // Finally call the actual method
                returnValue = originalMethod.apply(this, args);
                callDuration = getTimestamp() - timestampBefore;
            }
            finally {
                instanceMetadata.callDepth--;
            }
            // Restore unchanged properties
            restoreUnchangedProperties(this, originalProperties, cloneDepth);
        }, args, this);
        if (mutations) {
            // Throw when a method changes "this.prop.otherProp", but not "this.prop"
            if (mutations.this.some(change => change.path.length > 1) || Object.keys(mutations.args).length > 0) {
                // A method has changed a property without cloning the object first
                throw new MethodNotImmutableError(mutations, originalMethod, originalClass);
            }
        }
        if (classMetadata.settings.deepFreeze) {
            for (let key of objectKeys(this)) {
                const propertyValue = this[key];
                if (propertyValue !== originalProperties[key] && typeof propertyValue === 'object' && propertyValue) {
                    deepFreeze(propertyValue);
                }
            }
        }
        // Notify all observers added by observeImmutable
        const observers = instanceMetadata.observers;
        if (observers && observers.length > 0 && !instanceMetadata.callDepth) {
            const call = {
                instance: this,
                methodName,
                arguments: args,
                returnValue,
                callDuration,
                changes: undefined,
                oldProperties: originalProperties,
                newProperties: objectAssign({}, this)
            };
            let hasChanges = false;
            let changes = {};
            for (let key of objectKeys(this)) {
                if (this[key] !== originalProperties[key]) {
                    hasChanges = true;
                    changes[key] = {
                        oldValue: originalProperties[key],
                        newValue: this[key]
                    };
                }
            }
            if (hasChanges) {
                call.changes = changes;
            }
            for (let observer of observers) {
                observer(call);
            }
        }
        return returnValue;
    };
    return wrapped;
}
/** @internal */
export function restoreUnchangedProperties(target, original, depth) {
    for (let key of objectKeys(target)) {
        if (typeof target[key] === 'object' && target[key] !== null) {
            const propDepth = typeof depth === 'number' ? depth : (depth[key] || 0);
            if (propDepth > 0) {
                restoreUnchangedProperties(target[key], original[key], propDepth - 1);
                if (target[key] !== original[key] && flatEqual(target[key], original[key])) {
                    target[key] = original[key];
                }
            }
        }
    }
}
/** @internal */
export function isConstructingImmutable() {
    return constructing;
}
/** @internal */
function getTimestamp() {
    if (typeof performance === 'object') {
        // Cut values like 1234.8850000001 to 1234.885
        return (performance.now() * 1000 | 0) * 0.001;
    }
    else if (Date.now) {
        return Date.now();
    }
    else {
        return +(new Date());
    }
}

//# sourceMappingURL=immutable-decorator.js.map
