import { createObservable, observeImmutable } from './observe-immutable';
import { hasOwnProperty, objectKeys } from './utils';
/**
 * Implements an observable store with separate state action branches.
 */
export class ImmutableStateStore {
    constructor(actions) {
        this.actions = actions;
        this.observers = [];
        this.callObservers = [];
        this.subscriptions = [];
        this.combineInitialState();
        this.observeActionMethodCalls();
    }
    get state() {
        return this.currentState;
    }
    /** Remove all observers and free resources. */
    destroy() {
        this.observers.splice(0, this.observers.length);
        this.callObservers.splice(0, this.callObservers.length);
        this.subscriptions.forEach(s => s.unsubscribe());
    }
    observeCalls(observableClass) {
        return createObservable(this.callObservers, observableClass);
    }
    observeState(observableClass) {
        return createObservable(this.observers, observableClass);
    }
    /** Fully replace the application state (e.g. for undo/redo) */
    replaceState(newState) {
        const changes = {};
        for (let key of objectKeys(newState)) {
            if (this.currentState[key] !== newState[key]) {
                changes[key] = {
                    oldValue: this.currentState[key],
                    newValue: newState[key]
                };
            }
        }
        if (Object.keys(changes).length > 0) {
            this.propertiesChanged(undefined, changes);
            this.emitChangedState();
        }
    }
    /** Should only be used in testing. Use action methods to change the application state. */
    setStateForTesting(newState) {
        const changes = {};
        for (let key of objectKeys(newState)) {
            changes[key] = {
                oldValue: this.currentState[key],
                newValue: newState[key]
            };
        }
        this.propertiesChanged(undefined, changes);
        this.emitChangedState();
    }
    combineInitialState() {
        // Combine initial state properties of all branches into one object.
        const initialState = {};
        for (let branchName in this.actions) {
            const actionBranch = this.actions[branchName];
            for (let propKey of objectKeys(actionBranch)) {
                if (actionBranch[propKey] !== undefined || !(propKey in initialState)) {
                    initialState[propKey] = actionBranch[propKey];
                }
            }
        }
        this.currentState = initialState;
        // Apply initial state to all action branches
        for (let branchName in this.actions) {
            const actionBranch = this.actions[branchName];
            for (let propKey of objectKeys(actionBranch)) {
                actionBranch[propKey] = initialState[propKey];
            }
        }
    }
    observeActionMethodCalls() {
        // Observe all method calls of any action branch
        for (let branchName in this.actions) {
            const actionBranch = this.actions[branchName];
            const sub = observeImmutable(actionBranch).subscribe((call) => {
                const stateBefore = this.currentState;
                if (call.changes) {
                    this.propertiesChanged(call.instance, call.changes);
                    this.emitChangedState();
                }
                this.emitMethodCall(call, stateBefore);
            });
            this.subscriptions.push(sub);
        }
    }
    // Apply changes to this.currentState and all action branches that use the changed property
    propertiesChanged(instance, changes) {
        const newState = Object.assign({}, this.state);
        for (let changedProp in changes) {
            const newValue = changes[changedProp].newValue;
            newState[changedProp] = newValue;
            for (let branchToUpdateName in this.actions) {
                let branchToUpdate = this.actions[branchToUpdateName];
                if (branchToUpdate != instance && hasOwnProperty(branchToUpdate, changedProp)) {
                    branchToUpdate[changedProp] = newValue;
                }
            }
        }
        this.currentState = newState;
    }
    emitChangedState() {
        for (let observer of this.observers) {
            observer(this.currentState);
        }
    }
    emitMethodCall(actionMethodCall, stateBefore) {
        if (this.callObservers.length) {
            // Emit a tracked method call on "this".
            const actionName = Object.keys(this.actions)
                .filter(propertyKey => this.actions[propertyKey] === actionMethodCall.instance)[0];
            const callToEmit = {
                instance: this,
                methodName: actionName + '.' + actionMethodCall.methodName,
                arguments: actionMethodCall.arguments,
                returnValue: actionMethodCall.returnValue,
                callDuration: actionMethodCall.callDuration,
                changes: actionMethodCall.changes,
                oldProperties: stateBefore,
                newProperties: this.currentState,
            };
            for (let observer of this.callObservers) {
                observer(callToEmit);
            }
        }
    }
    handleSubscription(subscriber) {
        const callback = subscriber.next
            ? (newState) => subscriber.next(newState)
            : subscriber;
        this.observers.push(callback);
        return {
            unsubscribe: () => {
                let index;
                while ((index = this.observers.indexOf(callback)) >= 0) {
                    this.observers.splice(index, 1);
                }
            }
        };
    }
}

//# sourceMappingURL=immutable-state-store.js.map
